import unittest

from src.keewee import keewee, KeeWee


class ClosureKeeWee(unittest.TestCase):
    def setUp(self) -> None:
        self.x, self.set_x = keewee(name="x", initial=1)

    def test_dunder_eq(self):
        self.assertEqual(self.x, 1)

    def test_dunder_repr(self):
        self.assertEqual(repr(self.x), "1")

    def test_mutating(self):
        self.set_x(5)
        self.assertEqual(self.x, 5)

        KeeWee.pprint()


if __name__ == '__main__':
    unittest.main()



import random
import logging

from dataclasses import dataclass

logging.basicConfig(level=logging.DEBUG)


@dataclass
class Worker:
    name: str
    w_id: str
    skill_lvl: int


def business_logic(worker: Worker):
    for _ in range(10):
        worker.skill_lvl = random.randint(1, 10)


def main():
    logging.info(msg="  +-----------------------------------------------------+")
    logging.info(msg="  Initialize first worker")
    #
    worker = Worker(name="Worker1", w_id="1", skill_lvl=0)
    #
    logging.info(msg=f"    {worker}")
    logging.info(msg="  Calling business logic from within main function")
    #
    business_logic(worker)
    #
    logging.info(msg="  Business logic has ended in main function")
    logging.info(msg=f"    {worker}")
    logging.info(msg="  +-----------------------------------------------------+")


if __name__ == "__main__":
    main()



import unittest
from dataclasses import dataclass, field

from src.keewee import KeeWee


@dataclass
class PokemonTrainerSum:
    name: str
    skill_level: int | KeeWee = field(default=KeeWee(mode='sum'), repr=False)


@dataclass
class PokemonTrainerMax:
    name: str
    skill_level: int | KeeWee = field(default=KeeWee(mode='max'), repr=False)


@dataclass
class PokemonTrainerMin:
    name: str
    skill_level: int | KeeWee = field(default=KeeWee(mode='min'), repr=False)


class TestModes(unittest.TestCase):

    def test_mode_sum(self):
        pokemon_trainer = PokemonTrainerSum(name='Ash Ketchum', skill_level=0)
        pokemon_trainer.skill_level = 1
        self.assertEqual(KeeWee.dumpd().get("PokemonTrainerSum").get("skill_level").get("PokemonTrainerSum(name='Ash Ketchum')"), 1)

    def test_mode_max(self):
        pokemon_trainer = PokemonTrainerMax(name='Ash Ketchum', skill_level=0)
        pokemon_trainer.skill_level = 1
        self.assertEqual(KeeWee.dumpd().get("PokemonTrainerMax").get("skill_level").get("PokemonTrainerMax(name='Ash Ketchum')"), 1)
        pokemon_trainer.skill_level = 99
        self.assertEqual(KeeWee.dumpd().get("PokemonTrainerMax").get("skill_level").get("PokemonTrainerMax(name='Ash Ketchum')"), 99)

    def test_mode_min(self):
        pokemon_trainer = PokemonTrainerMin(name='Ash Ketchum', skill_level=0)
        pokemon_trainer.skill_level = 1
        self.assertEqual(KeeWee.dumpd().get("PokemonTrainerMin").get("skill_level").get("PokemonTrainerMin(name='Ash Ketchum')"), 0)
        pokemon_trainer.skill_level = -1
        self.assertEqual(KeeWee.dumpd().get("PokemonTrainerMin").get("skill_level").get("PokemonTrainerMin(name='Ash Ketchum')"), -1)


if __name__ == '__main__':
    unittest.main()



import unittest

from src.keewee import KeeWee


class Employee:
    skill_level = KeeWee(mode="list")
    kids = KeeWee(mode="dtv")

    def __init__(self, name: str, skill_level: int):
        self.name = name
        self.skill_level = skill_level

    def __str__(self):
        return f"{self.name})"


class MyTestCase(unittest.TestCase):
    def test_something(self):
        andrew = Employee(name="Andrea Taylor", skill_level=0)
        self.assertEqual(andrew.skill_level, 0)
        andrew.skill_level = 1
        andrew.skill_level = 1
        andrew.skill_level = 1
        andrew.skill_level = 3
        andrew.skill_level = 1
        self.assertEqual(andrew.skill_level, 1)

        andrew.kids = 1
        andrew.kids = 2
        KeeWee.pprint()


if __name__ == "__main__":
    unittest.main()



"""Implements an autovivification dictionary for global state recording and management"""
from __future__ import annotations

import datetime as dt
import inspect
import json
import uuid

from pprint import pp
from typing import overload, cast, Any, Callable, TypeAlias

ISOFormat: TypeAlias = str


class KeeWeeRepo(dict):
    """Global application state repository for recording and management.
    Implements the missing-dunder method for autovivification."""

    def __missing__(self, kee):
        wee = self[kee] = type(self)()
        return wee


KEEWEE_REPO = KeeWeeRepo()


def rec_mode_direct(kw_store: dict[str, set], key: str, value: Any) -> None:
    """default mode that sets a value for a key into a dictionary

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: any value
    """
    kw_store[key] = value


def rec_mode_set(kw_store: dict[str, set], key: str, value: Any) -> None:
    """Collects the occurring values in a set.

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: any value added to a set
    """
    if not kw_store.get(key):
        kw_store[key] = set()
    kw_store[key].add(value)


def rec_mode_list(kw_store: dict[str, list], key: str, value: Any) -> None:
    """Collects the occurring values in a list.

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: any value added to a list
    """
    if not kw_store.get(key):
        kw_store[key] = []
    kw_store[key].append(value)


def rec_mode_dtv(kw_store: dict[str, dict[ISOFormat, Any]], key: str, value: Any) -> None:
    """Collects the occurring values in dictionary that maps the current timestamp
    in ISO-format to the new value

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: any value added to a mapping of timestamp to value
    """
    if not kw_store.get(key):
        kw_store[key] = {}
    kw_store[key][dt.datetime.now().time().isoformat()] = value


def rec_mode_sum(kw_store: dict[str, int | float], key: str, value: int | float) -> None:
    """Sums al occurring values

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: numerical value
    """
    if kw_store.get(key) is None:
        kw_store[key] = value
    kw_store[key] += value


def rec_mode_max(kw_store: dict[str, int | float], key: str, value: int | float) -> None:
    """Stores the maximum of all occurring values

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: numerical value
    """
    if kw_store.get(key) is None:
        kw_store[key] = value
    kw_store[key] = max(kw_store.get(key), value)


def rec_mode_min(kw_store: dict[str, int | float], key: str, value: int | float) -> None:
    """Stores the minimum of all occurring values

    :param kw_store: The global KeeWee repository
    :param key: a string key
    :param value: numerical value
    """
    if kw_store.get(key) is None:
        kw_store[key] = value
    kw_store[key] = min(kw_store.get(key), value)


RECORD_MODES: dict[str, Callable[..., None]] = {
    "direct": rec_mode_direct,
    "list": rec_mode_list,
    "set": rec_mode_set,
    "dtv": rec_mode_dtv,  # iso-format
    "sum": rec_mode_sum,
    "max": rec_mode_max,
    "min": rec_mode_min,
}


class KeeWee:
    """KeeWee implements the descriptor-protocol and hooks into a specific
    recording method for keeping the state
    """
    _repo = KEEWEE_REPO

    def __init__(self, mode: str = "list", blame: bool = False):
        self.blame = blame
        self.record_mode = RECORD_MODES.get(mode, RECORD_MODES['list'])
        self.public_name = ""
        self.private_name = ""

    def __set_name__(self, owner: type[object], name: str) -> None:
        """Does not seem to change over the mode"""
        self.public_name = name
        self.private_name = '_' + name

    def __set__(self, obj: object, value: int) -> None:
        obj.__dict__[self.private_name] = value
        owner = obj.__class__.__name__
        instance_name = str(obj)
        if self.blame:
            blamed_value: tuple[int, str] = (value, inspect.getouterframes(inspect.currentframe(), 2)[1][3])
            self.record_mode(kw_store=self._repo[owner][self.public_name], key=instance_name, value=blamed_value)
        else:
            self.record_mode(kw_store=self._repo[owner][self.public_name], key=instance_name, value=value)

    @overload
    def __get__(self, obj: None, obj_type: None) -> KeeWee:
        ...

    @overload
    def __get__(self, obj: object, obj_type: type[object]) -> int:
        ...

    def __get__(self, obj: object | None, obj_type: type[object] | None = None) -> KeeWee | int:
        """Does not seem to change over the mode"""
        if obj is None:
            return self
        return cast(int, obj.__dict__.get(self.private_name))

    @classmethod
    def pprint(cls) -> None:
        pp(cls._repo)

    @classmethod
    def dump(cls, file_name: str):
        with open(file_name, "w", encoding="utf-8") as f:
            json.dump(cls._repo, f)

    @classmethod
    def dumps(cls):
        return json.dumps(cls._repo)

    @classmethod
    def dumpd(cls):
        return cls._repo


def keewee(initial: Any, name: str | None = None, mode: str = "direct"):
    if mode != "direct":
        raise NotImplementedError("Another mode is currently not available for the keewee-hook!")
    name = name if name else uuid.uuid4().hex
    rec_mode = RECORD_MODES.get(mode)

    def setter(value):
        rec_mode(KEEWEE_REPO, name, value)

    setter(initial)

    class IntProxy(int):
        def __new__(cls, value, *args, **kwargs):
            return super(cls, cls).__new__(cls, KEEWEE_REPO.get(name))

        def __int__(self):
            return KEEWEE_REPO.get(name)

        def __repr__(self) -> str:
            return f"{int(self)}"

        def __eq__(self, __o: object) -> bool:
            return int(self) == __o

    match initial:
        case int():
            return IntProxy(initial), setter



import json
import pprint
import importlib
import inspect
import logging

from dataclasses import dataclass, field

from src import KeeWee

logging.basicConfig(level=logging.DEBUG)


def run_from():
    with open("keewee.config.json", "r") as f:
        kw = json.load(f)
    testmain = importlib.import_module(kw["module"])
    print(testmain)
    print(testmain["Worker"])


def monkey_patch():
    import some_app
    DC_ATT_STR = "{}"
    DC_FIELD_STR = "dataclasses.field(default=KeeWee(), repr=False)"
    patch_str = "    skill_lvl: int | KeeWee = field(default=KeeWee(), repr=False)\n"
    source_worker, _ = inspect.getsourcelines(some_app.Worker)
    source_worker[4] = patch_str
    exec("".join(source_worker))
    some_app.Worker = locals()["Worker"]
    logging.info(msg="  ")
    some_app.main()
    logging.info(msg="  ")

    logging.info(msg=pprint.pformat(KeeWee.dumpd()))


if __name__ == "__main__":
    monkey_patch()
    # run_from()



import unittest

from dataclasses import dataclass, field
from random import randint

from src.keewee import KeeWee


@dataclass
class Employee:
    id: str
    name: str
    position: str | KeeWee = field(default=KeeWee(mode="dtv"), repr=False)
    skill_lvl: int | KeeWee = field(default=KeeWee(mode="list"), repr=False)

    def __repr__(self):
        return f"<{self.id}|{self.name}>"


class TestModeList(unittest.TestCase):

    def setUp(self) -> None:
        self.e1 = Employee(id="1", name="First", position="CEO", skill_lvl=0)
        self.e2 = Employee(id="2", name="Second", position="STUDENT", skill_lvl=0)

    def test_something(self):
        self.e1.position = "COO"
        self.e1.position = "COO"
        self.e1.position = "COO"
        self.e1.position = "COO"
        self.e1.position = "COOL"
        self.e2.position = "WERKSTUDENT"
        for _ in range(5):
            self.e2.skill_lvl = randint(1, 10)
        KeeWee.pprint()


if __name__ == '__main__':
    unittest.main()
